from __future__ import annotations

from collections.abc import Awaitable, Callable, Collection, Iterable, Mapping, Sequence
from functools import partial, wraps
from itertools import chain, compress, filterfalse, repeat, starmap
from pathlib import Path
from typing import Literal, Never, TypedDict

from yarl import URL

from . import config_ctx, sync_ctx
from ._utils.aio import gather, run_in_thread
from ._utils.attrs import evolve
from ._utils.file import trash
from ._utils.iteration import bucketise, uniq
from .definitions import Defn, Strategy
from .pkg_archives._download import download_pkg_archive
from .pkg_db import Connection, Row, transact, use_tuple_factory
from .pkg_db.models import Pkg, PkgLoggedVersion, make_db_converter
from .progress_reporting import make_incrementing_progress_tracker
from .resolvers import PkgCandidate
from .results import (
    AnyResult,
    ManagerError,
    PkgAlreadyInstalled,
    PkgConflictsWithInstalled,
    PkgConflictsWithUnreconciled,
    PkgFilesNotMatching,
    PkgInstalled,
    PkgNotInstalled,
    PkgRemoved,
    PkgStrategiesUnsupported,
    PkgUpdated,
    PkgUpToDate,
    is_error_result,
    resultify,
)

_MUTATE_PKGS_LOCK = '_MUTATE_PKGS_'


_download_pkg_archive = resultify(download_pkg_archive)


def _with_mutate_lock[**P, T](
    coro_fn: Callable[P, Awaitable[T]],
) -> Callable[P, Awaitable[T]]:
    @wraps(coro_fn)
    async def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        async with sync_ctx.locks()[_MUTATE_PKGS_LOCK, config_ctx.config().profile]:
            return await coro_fn(*args, **kwargs)

    return inner


def split_results[T](
    values: Iterable[tuple[Defn, AnyResult[T]]],
) -> tuple[dict[Defn, T], dict[Defn, AnyResult[Never]]]:
    ts: dict[Defn, T] = {}
    errors: dict[Defn, AnyResult[Never]] = {}

    for defn, value in values:
        if is_error_result(value):
            errors[defn] = value
        else:
            ts[defn] = value

    return ts, errors


def get_alias_from_url(value: str) -> tuple[str, str] | None:
    "Attempt to extract a valid ``Defn`` source and alias from a URL."
    resolvers = config_ctx.resolvers()
    url = URL(value)
    return next(
        ((r.metadata.id, a) for r in resolvers.values() if (a := r.get_alias_from_url(url))),
        None,
    )


async def get_changelog(source: str, uri: str) -> str:
    "Retrieve a changelog from a URI."
    return await config_ctx.resolvers().get_or_dummy(source).get_changelog(URL(uri))


def build_pkg_from_pkg_candidate(
    defn: Defn,
    pkg_candidate: PkgCandidate,
    *,
    folders: list[TypedDict[{'name': str}]],
) -> Pkg:
    return make_db_converter().structure(
        {
            'deps': [],
        }
        | pkg_candidate
        | {
            'source': defn.source,
            'options': {k: bool(v) for k, v in defn.strategies.items()},
            'folders': folders,
        },
        Pkg,
    )


def build_pkg_from_row_mapping(connection: Connection, row_mapping: Row) -> Pkg:
    fk = {'pkg_source': row_mapping['source'], 'pkg_id': row_mapping['id']}
    return make_db_converter().structure(
        {
            **row_mapping,
            'options': connection.execute(
                """
                SELECT any_flavour, any_release_type, version_eq
                FROM pkg_options
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchone(),
            'folders': connection.execute(
                """
                SELECT name
                FROM pkg_folder
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchall(),
            'deps': connection.execute(
                """
                SELECT id
                FROM pkg_dep
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                """,
                fk,
            ).fetchall(),
        },
        Pkg,
    )


def _check_pkgs_not_exist(defns: Collection[Defn]) -> list[bool]:
    "Check that packages exist in the database."
    if not defns:
        return []

    with config_ctx.database() as connection, use_tuple_factory(connection) as cursor:
        return [
            e
            for (e,) in cursor.execute(
                f"""
                WITH defn (source, alias, id)
                AS (
                    VALUES {', '.join(('(?, ?, ?)',) * len(defns))}
                )
                SELECT NOT EXISTS (
                    SELECT 1
                    FROM pkg
                    WHERE pkg.source = defn.source AND (
                        pkg.id = defn.alias OR pkg.id = defn.id OR lower(pkg.slug) = lower(defn.alias)
                    )
                )
                FROM defn
                """,
                tuple(i for d in defns for i in (d.source, d.alias, d.id)),
            ).fetchall()
        ]


def get_pkgs(defns: Collection[Defn] | Literal['all']) -> list[Pkg | None]:
    if defns != 'all' and not defns:
        return []

    with config_ctx.database() as connection:
        if defns == 'all':
            pkgs = connection.execute(
                """
                SELECT * FROM pkg
                """,
            ).fetchall()
        else:
            pkgs = connection.execute(
                f"""
                WITH defn (source, alias, id)
                AS (
                    VALUES {', '.join(('(?, ?, ?)',) * len(defns))}
                )
                SELECT pkg.*
                FROM defn
                LEFT JOIN pkg ON pkg.source = defn.source AND (
                    pkg.id = defn.alias OR pkg.id = defn.id OR lower(pkg.slug) = lower(defn.alias)
                )
                """,
                tuple(i for d in defns for i in (d.source, d.alias, d.id)),
            ).fetchall()

        return [build_pkg_from_row_mapping(connection, m) if m['source'] else None for m in pkgs]


def get_pinnable_pkgs(
    defns: Collection[Defn],
) -> list[AnyResult[Pkg]]:
    resolvers = config_ctx.resolvers()

    @resultify
    def validate_pkg(defn: Defn, pkg: Pkg | None):
        if Strategy.VersionEq not in resolvers.get_or_dummy(defn.source).metadata.strategies:
            raise PkgStrategiesUnsupported({Strategy.VersionEq})

        if not pkg:
            raise PkgNotInstalled()

        else:
            version = defn.strategies[Strategy.VersionEq]
            if version and pkg.version != version:
                raise PkgFilesNotMatching(defn.strategies)

        return pkg

    return list(map(validate_pkg, defns, get_pkgs(defns)))


def get_pkg_logged_versions(pkg: Pkg) -> list[PkgLoggedVersion]:
    convert = partial(make_db_converter().structure, cl=PkgLoggedVersion)

    with config_ctx.database() as connection:
        return [
            convert(v)
            for v in connection.execute(
                """
                SELECT version, install_time
                FROM pkg_version_log
                WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
                ORDER BY install_time DESC
                LIMIT 10
                """,
                {'pkg_source': pkg.source, 'pkg_id': pkg.id},
            ).fetchall()
        ]


def _insert_pkg(pkg: Pkg, transaction: Connection) -> None:
    pkg_values = make_db_converter().unstructure(pkg)

    transaction.execute(
        """
        INSERT INTO pkg (
            source,
            id,
            slug,
            name,
            description,
            url,
            download_url,
            date_published,
            version,
            changelog_url
        )
        VALUES (
            :source,
            :id,
            :slug,
            :name,
            :description,
            :url,
            :download_url,
            :date_published,
            :version,
            :changelog_url
        )
        """,
        pkg_values,
    )
    transaction.execute(
        """
        INSERT INTO pkg_options (
            any_flavour,
            any_release_type,
            version_eq,
            pkg_source,
            pkg_id
        )
        VALUES (
            :any_flavour,
            :any_release_type,
            :version_eq,
            :pkg_source,
            :pkg_id
        )
        """,
        pkg_values['options'] | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']},
    )
    transaction.executemany(
        """
        INSERT INTO pkg_folder (
            name,
            pkg_source,
            pkg_id
        )
        VALUES (
            :name,
            :pkg_source,
            :pkg_id
        )
        """,
        [
            f | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']}
            for f in pkg_values['folders']
        ],
    )
    if pkg_values['deps']:
        transaction.executemany(
            """
            INSERT INTO pkg_dep (
                id,
                pkg_source,
                pkg_id
            )
            VALUES (
                :id,
                :pkg_source,
                :pkg_id
            )
            """,
            [
                f | {'pkg_source': pkg_values['source'], 'pkg_id': pkg_values['id']}
                for f in pkg_values['deps']
            ],
        )
    transaction.execute(
        """
        INSERT OR IGNORE INTO pkg_version_log (
            version,
            pkg_source,
            pkg_id
        )
        VALUES (
            :version,
            :pkg_source,
            :pkg_id
        )
        """,
        {
            'version': pkg_values['version'],
            'pkg_source': pkg_values['source'],
            'pkg_id': pkg_values['id'],
        },
    )


def _delete_pkg(pkg: Pkg, transaction: Connection) -> None:
    transaction.execute(
        'DELETE FROM pkg WHERE source = :source AND id = :id',
        {
            'source': pkg.source,
            'id': pkg.id,
        },
    )


async def find_equivalent_pkg_defns(
    pkgs: Collection[Pkg],
) -> dict[Pkg, list[Defn]]:
    "Given a list of packages, find ``Defn``s of each package from other sources."
    from .catalogue import synchronise as synchronise_catalogue
    from .matchers import AddonFolder
    from .wow_installations import to_flavour

    config = config_ctx.config()
    resolvers = config_ctx.resolvers()
    flavour = to_flavour(config.track)

    catalogue = await synchronise_catalogue()

    @run_in_thread
    def collect_addon_folders():
        return {
            p: frozenset(
                a
                for f in p.folders
                for a in (AddonFolder.from_path(flavour, config.addon_dir / f.name),)
                if a
            )
            for p in pkgs
        }

    def get_catalogue_defns(pkg: Pkg) -> frozenset[Defn]:
        entry = catalogue.keyed_entries.get((pkg.source, pkg.id))
        return frozenset(Defn(s.source, s.id) for s in entry.same_as) if entry else frozenset()

    def get_addon_toc_defns(pkg_source: str, addon_folders: Collection[AddonFolder]):
        return frozenset(
            d
            for a in addon_folders
            for d in a.get_defns_from_toc_keys(resolvers.addon_toc_key_and_id_pairs)
            if d.source != pkg_source
        )

    folders_per_pkg = await collect_addon_folders()

    return {
        p: sorted(d, key=lambda d: resolvers.priority_dict[d.source])
        for p in pkgs
        for d in (get_catalogue_defns(p) | get_addon_toc_defns(p.source, folders_per_pkg[p]),)
        if d
    }


async def _resolve_deps(
    results: Mapping[Defn, AnyResult[PkgCandidate]],
) -> Mapping[Defn, AnyResult[PkgCandidate]]:
    """Resolve package dependencies.

    The resolver will not follow dependencies
    more than one level deep.  This is to avoid unnecessary
    complexity for something that I would never expect to
    encounter in the wild.
    """
    pkg_candidates, _ = split_results(results.items())
    dep_defns = uniq(
        filterfalse(
            {(d.source, p['id']) for d, p in pkg_candidates.items()}.__contains__,
            (
                (d.source, e['id'])
                for d, p in pkg_candidates.items()
                if 'deps' in p
                for e in p['deps']
            ),
        )
    )
    if not dep_defns:
        return {}

    # Map the ID both to the `alias` and the `id` fields of the `Defn` so that
    # it's not lost if we humanise the alias later.
    deps = await resolve(list(starmap(Defn, *zip((s, i, i) for s, i in dep_defns))))
    pretty_deps = {
        evolve(d, {'alias': r['slug']}) if isinstance(r, dict) else d: r for d, r in deps.items()
    }
    return pretty_deps


async def resolve(
    defns: Collection[Defn], with_deps: bool = False
) -> Mapping[Defn, AnyResult[PkgCandidate]]:
    "Resolve definitions into packages."
    if not defns:
        return {}

    resolvers = config_ctx.resolvers()

    defns_by_source = bucketise(defns, key=lambda v: v.source)

    track_progress = make_incrementing_progress_tracker(len(defns_by_source), 'Resolving add-ons')

    results = await gather(
        track_progress(resultify(resolvers.get_or_dummy(s).resolve)(b))
        for s, b in defns_by_source.items()
    )
    results_by_defn = dict(
        chain(
            zip(defns, repeat(ManagerError())),
            *(
                r.items() if isinstance(r, dict) else zip(d, repeat(r))
                for d, r in zip(defns_by_source.values(), results)
            ),
        )
    )
    if with_deps:
        results_by_defn |= await _resolve_deps(results_by_defn)

    return results_by_defn


@resultify
@run_in_thread
def _mutate_install(
    defn: Defn, pkg_candidate: PkgCandidate, archive: Path, *, replace_folders: bool
):
    with (
        config_ctx.resolvers().archive_opener_dict[defn.source](
            archive,
        ) as (top_level_folders, extract),
        config_ctx.database() as connection,
    ):
        installed_conflicts = connection.execute(
            f"""
            SELECT DISTINCT pkg.*
            FROM pkg
            JOIN pkg_folder ON pkg_folder.pkg_source = pkg.source AND pkg_folder.pkg_id = pkg.id
            WHERE pkg_folder.name IN ({', '.join(('?',) * len(top_level_folders))})
            """,
            tuple(top_level_folders),
        ).fetchall()
        if installed_conflicts:
            raise PkgConflictsWithInstalled(installed_conflicts)

        config = config_ctx.config()

        if replace_folders:
            trash(config.addon_dir / f for f in top_level_folders)
        else:
            unreconciled_conflicts = top_level_folders & {
                f.name for f in config.addon_dir.iterdir()
            }
            if unreconciled_conflicts:
                raise PkgConflictsWithUnreconciled(unreconciled_conflicts)

        extract(config.addon_dir)

        pkg = build_pkg_from_pkg_candidate(
            defn, pkg_candidate, folders=[{'name': f} for f in sorted(top_level_folders)]
        )
        with transact(connection) as transaction:
            _insert_pkg(pkg, transaction)

    return PkgInstalled(pkg)


@resultify
@run_in_thread
def _mutate_update(defn: Defn, old_pkg: Pkg, pkg_candidate: PkgCandidate, archive: Path):
    with (
        config_ctx.resolvers().archive_opener_dict[defn.source](
            archive,
        ) as (top_level_folders, extract),
        config_ctx.database() as connection,
    ):
        installed_conflicts = connection.execute(
            f"""
            SELECT DISTINCT pkg.*
            FROM pkg
            JOIN pkg_folder ON pkg_folder.pkg_source = pkg.source AND pkg_folder.pkg_id = pkg.id
            WHERE (pkg_folder.pkg_source != ? AND pkg_folder.pkg_id != ?)
                AND (pkg_folder.name IN ({', '.join(('?',) * len(top_level_folders))}))
            """,
            (defn.source, pkg_candidate['id'], *top_level_folders),
        ).fetchall()
        if installed_conflicts:
            raise PkgConflictsWithInstalled(installed_conflicts)

        config = config_ctx.config()

        unreconciled_conflicts = top_level_folders - {f.name for f in old_pkg.folders} & {
            f.name for f in config.addon_dir.iterdir()
        }
        if unreconciled_conflicts:
            raise PkgConflictsWithUnreconciled(unreconciled_conflicts)

        trash(config.addon_dir / f.name for f in old_pkg.folders)
        extract(config.addon_dir)

        new_pkg = build_pkg_from_pkg_candidate(
            defn, pkg_candidate, folders=[{'name': f} for f in sorted(top_level_folders)]
        )
        with transact(connection) as transaction:
            _delete_pkg(old_pkg, transaction)
            _insert_pkg(new_pkg, transaction)

    return PkgUpdated(old_pkg, new_pkg)


@resultify
@run_in_thread
def _mutate_remove(defn: Defn, pkg: Pkg, *, keep_folders: bool):
    if not keep_folders:
        config = config_ctx.config()
        trash(config.addon_dir / f.name for f in pkg.folders)

    with config_ctx.database() as connection, transact(connection) as transaction:
        _delete_pkg(pkg, transaction)

    return PkgRemoved(pkg)


@resultify
def _mutate_pin(defn: Defn, pkg: Pkg):
    with config_ctx.database() as connection, transact(connection) as transaction:
        (version_eq,) = transaction.execute(
            """
            UPDATE pkg_options
            SET version_eq = :version_eq
            WHERE pkg_source = :pkg_source AND pkg_id = :pkg_id
            RETURNING version_eq
            """,
            {
                'pkg_source': pkg.source,
                'pkg_id': pkg.id,
                'version_eq': defn.strategies[Strategy.VersionEq] is not None,
            },
        ).fetchone()

    return PkgInstalled(
        evolve(pkg, {'options': {'version_eq': bool(version_eq)}}),
    )


@run_in_thread
def _check_installed_pkg_integrity(addon_dir: Path, pkg: Pkg):
    return all((addon_dir / p.name).exists() for p in pkg.folders)


@_with_mutate_lock
async def install(
    defns: Sequence[Defn],
    *,
    replace_folders: bool,
    dry_run: bool = False,
) -> Mapping[Defn, AnyResult[PkgInstalled]]:
    "Install packages from a definition list."

    # We'll weed out installed deps from the results after resolving -
    # doing it this way isn't particularly efficient but avoids having to
    # deal with local state in ``resolve``.
    resolve_results = await resolve(
        list(compress(defns, _check_pkgs_not_exist(defns))), with_deps=True
    )
    pkg_candidates, resolve_errors = split_results(resolve_results.items())
    pkg_candidates = dict(
        compress(
            pkg_candidates.items(),
            _check_pkgs_not_exist([evolve(d, {'id': c['id']}) for d, c in pkg_candidates.items()]),
        )
    )

    results = dict.fromkeys(defns, PkgAlreadyInstalled()) | resolve_errors

    if dry_run:
        return results | {
            d: PkgInstalled(build_pkg_from_pkg_candidate(d, p, folders=[]), dry_run=True)
            for d, p in pkg_candidates.items()
        }

    download_results = await gather(
        _download_pkg_archive(d, r['download_url']) for d, r in pkg_candidates.items()
    )
    archive_paths, download_errors = split_results(zip(pkg_candidates, download_results))

    track_progress = make_incrementing_progress_tracker(len(archive_paths), 'Installing')

    return (
        results
        | download_errors
        | {
            d: await track_progress(
                _mutate_install(d, pkg_candidates[d], a, replace_folders=replace_folders)
            )
            for d, a in archive_paths.items()
        }
    )


@_with_mutate_lock
async def replace(
    defns: Mapping[Defn, Defn],
) -> Mapping[Defn, AnyResult[PkgInstalled | PkgRemoved]]:
    "Replace installed packages with re-reconciled packages."

    inverse_defns = {v: k for k, v in defns.items()}
    if len(inverse_defns) != len(defns):
        raise ValueError('``defns`` must be unique')

    old_pkgs = {d: p for d, p in zip(defns, get_pkgs(defns)) if p}

    resolve_results = await resolve(
        list(compress(defns.values(), _check_pkgs_not_exist(defns.values()))),
        with_deps=True,
    )
    pkg_candidates, resolve_errors = split_results(resolve_results.items())
    pkg_candidates = dict(
        compress(
            pkg_candidates.items(),
            _check_pkgs_not_exist([evolve(d, {'id': c['id']}) for d, c in pkg_candidates.items()]),
        )
    )

    results = dict(
        chain.from_iterable(
            zip(
                ((d, PkgNotInstalled()) for d in defns),
                ((d, None) for d in defns.values()),
                strict=True,
            )
        )
    )
    results = results | resolve_errors

    download_results = await gather(
        _download_pkg_archive(d, r['download_url']) for d, r in pkg_candidates.items()
    )
    archive_paths, download_errors = split_results(zip(pkg_candidates, download_results))

    replace_coros = {
        k: c
        for d, a in archive_paths.items()
        for o in (inverse_defns[d],)
        for k, c in (
            (o, _mutate_remove(d, old_pkgs[o], keep_folders=False)),
            (d, _mutate_install(d, pkg_candidates[d], a, replace_folders=False)),
        )
    }

    track_progress = make_incrementing_progress_tracker(len(replace_coros), 'Replacing')

    results = (
        results | download_errors | {k: await track_progress(c) for k, c in replace_coros.items()}
    )
    return {k: v for k, v in results.items() if v is not None}


@_with_mutate_lock
async def update(
    defns: Sequence[Defn] | Literal['all'],
    *,
    dry_run: bool = False,
) -> Mapping[Defn, AnyResult[PkgInstalled | PkgUpdated]]:
    "Update installed packages from a definition list."

    config = config_ctx.config()

    if defns == 'all':
        defns_to_pkgs = {p.to_defn(): p for p in get_pkgs(defns) if p}
        defns = list(defns_to_pkgs)
        resolve_defns = {d: d for d in defns_to_pkgs}

    else:
        defns_to_pkgs = {d: p for d, p in zip(defns, get_pkgs(defns)) if p}
        resolve_defns = {
            # Attach the source ID to each ``Defn`` from the
            # corresponding installed package.  Using the ID has the benefit
            # of resolving installed-but-renamed packages - the slug is
            # transient but the ID isn't
            evolve(d, {'id': p.id}) if d.strategies else p.to_defn(): d
            for d, p in defns_to_pkgs.items()
        }

    resolve_results = await resolve(resolve_defns, with_deps=True)
    pkg_candidates, resolve_errors = split_results(
        # Discard the reconstructed ``Defn``s
        (resolve_defns.get(d) or d, r)
        for d, r in resolve_results.items()
    )

    updatables = {
        d: (o, n)
        for d, n in pkg_candidates.items()
        for o in (defns_to_pkgs.get(d),)
        if not o
        or o.version != n['version']
        or not await _check_installed_pkg_integrity(config.addon_dir, o)
    }

    results = (
        dict.fromkeys(defns, PkgNotInstalled())
        | resolve_errors
        | {
            d: PkgUpToDate(is_pinned=bool(defns_to_pkgs[d].options.version_eq))
            for d in pkg_candidates.keys() - updatables.keys()
        }
    )

    if dry_run:
        return results | {
            d: PkgUpdated(o, build_pkg_from_pkg_candidate(d, n, folders=[]), dry_run=True)
            if o
            else PkgInstalled(build_pkg_from_pkg_candidate(d, n, folders=[]), dry_run=True)
            for d, (o, n) in updatables.items()
        }

    download_results = await gather(
        _download_pkg_archive(d, n['download_url']) for d, (_, n) in updatables.items()
    )
    archive_paths, download_errors = split_results(zip(updatables, download_results))

    track_progress = make_incrementing_progress_tracker(len(archive_paths), 'Updating')

    return (
        results
        | download_errors
        | {
            d: await track_progress(
                _mutate_update(d, o, n, a)
                if o
                else _mutate_install(d, n, a, replace_folders=False)
            )
            for d, a in archive_paths.items()
            for o, n in (updatables[d],)
        }
    )


@_with_mutate_lock
async def remove(
    defns: Sequence[Defn], *, keep_folders: bool
) -> Mapping[Defn, AnyResult[PkgRemoved]]:
    "Remove packages by their definition."
    return {
        d: await _mutate_remove(d, p, keep_folders=keep_folders) if p else PkgNotInstalled()
        for d, p in zip(defns, get_pkgs(defns))
    }


@_with_mutate_lock
async def pin(defns: Sequence[Defn]) -> Mapping[Defn, AnyResult[PkgInstalled]]:
    """Pin and unpin installed packages.

    instawow does not have true pinning.  This flips ``Strategy.VersionEq``
    on for installed packages from sources that support it.
    The net effect is the same as if the package
    had been reinstalled with the ``VersionEq`` strategy.
    """
    return {
        d: r if is_error_result(r) else _mutate_pin(d, r)
        for d, r in zip(defns, get_pinnable_pkgs(defns))
    }



from __future__ import annotations

from pathlib import Path

import pytest
from yarl import URL

from instawow import config_ctx, pkg_management
from instawow.definitions import Defn
from instawow.wow_installations import extract_installation_dir_from_addon_dir
from instawow_weakaura_updater.builder import (
    _Aura,
    _generate_addon,
    _WeakAuras,
    build_addon,
    extract_installed_auras,
)
from instawow_weakaura_updater.config import PluginConfig

pytestmark = pytest.mark.usefixtures('_iw_config_ctx')


@pytest.fixture
async def plugin_config():
    return PluginConfig(profile_config=config_ctx.config()).ensure_dirs()


@pytest.fixture
async def saved_vars_path():
    installation_path = extract_installation_dir_from_addon_dir(
        config_ctx.config().addon_dir,
    )
    assert installation_path
    saved_vars_path = installation_path.joinpath('WTF/Account/Instawow/SavedVariables')
    saved_vars_path.mkdir(parents=True)
    return saved_vars_path


def test_extract_no_saved_vars(
    plugin_config: PluginConfig,
):
    assert list(extract_installed_auras(plugin_config)) == []


def test_extract_empty_displays_table(
    plugin_config: PluginConfig,
    saved_vars_path: Path,
):
    (saved_vars_path / _WeakAuras.name).with_suffix('.lua').write_text(
        """\
WeakAurasSaved = {
    ["displays"] = {
    },
}
""",
        encoding='utf-8',
    )
    assert list(extract_installed_auras(plugin_config)) == [('Instawow', _WeakAuras, {})]


def test_extract_urlless_display(
    plugin_config: PluginConfig,
    saved_vars_path: Path,
):
    (saved_vars_path / _WeakAuras.name).with_suffix('.lua').write_text(
        """\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
""",
        encoding='utf-8',
    )
    assert list(extract_installed_auras(plugin_config)) == [('Instawow', _WeakAuras, {})]


def test_extract_url_host_not_wago_display(
    plugin_config: PluginConfig,
    saved_vars_path: Path,
):
    (saved_vars_path / _WeakAuras.name).with_suffix('.lua').write_text(
        """\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wafo.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
""",
        encoding='utf-8',
    )
    assert list(extract_installed_auras(plugin_config)) == [('Instawow', _WeakAuras, {})]


def test_extract_minimal_wago_display(
    plugin_config: PluginConfig,
    saved_vars_path: Path,
):
    (saved_vars_path / _WeakAuras.name).with_suffix('.lua').write_text(
        """\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
            ["url"] = "https://wago.io/foo/1",
            ["version"] = 1,
            ["id"] = "foo",
            ["uid"] = "foo",
        },
    },
}
""",
        encoding='utf-8',
    )
    assert list(extract_installed_auras(plugin_config)) == [
        (
            'Instawow',
            _WeakAuras,
            {
                'foo': [
                    _Aura(
                        id='foo',
                        uid='foo',
                        parent=None,
                        url=URL('https://wago.io/foo/1'),
                        version=1,
                    )
                ]
            },
        )
    ]


def test_build_addon_no_auras(
    plugin_config: PluginConfig,
):
    _generate_addon(plugin_config, {})


async def test_build_addon_no_wago_auras(
    plugin_config: PluginConfig,
    saved_vars_path: Path,
):
    (saved_vars_path / _WeakAuras.name).with_suffix('.lua').write_text(
        """\
WeakAurasSaved = {
    ["displays"] = {
        ["Foo"] = {
            ["bar"] = "baz",
        },
    },
}
""",
        encoding='utf-8',
    )

    await build_addon(plugin_config)


def test_generate_reproducible(
    plugin_config: PluginConfig,
):
    checksum = _generate_addon(plugin_config, {}).version.read_text(encoding='utf-8')
    assert checksum == _generate_addon(plugin_config, {}).version.read_text(encoding='utf-8')


def test_generate_emits_changelog(
    plugin_config: PluginConfig,
):
    assert _generate_addon(plugin_config, {}).changelog.read_text(encoding='utf-8') == 'n/a'


async def test_resolve_weakauras_companion_pkg(
    plugin_config: PluginConfig,
):
    await build_addon(plugin_config)
    defn = Defn('instawow', 'weakauras-companion')
    resolve_results = await pkg_management.resolve([defn])
    assert type(resolve_results[defn]) is dict


async def test_resolve_weakauras_companion_autoupdate_pkg():
    defn = Defn('instawow', 'weakauras-companion-autoupdate')
    resolve_results = await pkg_management.resolve([defn])
    assert type(resolve_results[defn]) is dict



from __future__ import annotations

import importlib.metadata

import aiohttp.web
import pytest

from instawow._version import get_version, is_outdated

from ._fixtures.http import AddRoutes, Route


def test_get_version_same_as_importlib():
    assert get_version() == importlib.metadata.version('instawow')


def test_get_dummy_version_when_metadata_missing(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr('sys.path', [])
    assert get_version() == '0+dev'


@pytest.mark.parametrize(
    '_iw_mock_aiohttp_requests',
    [set()],
    indirect=True,
)
@pytest.mark.usefixtures('_iw_web_client_ctx')
async def test_is_outdated_works_in_variety_of_scenarios(
    monkeypatch: pytest.MonkeyPatch,
    iw_add_routes: AddRoutes,
):
    # Endpoint not responsive, version not cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '0.1.0')
        iw_add_routes(
            Route(
                r'//pypi\.org/simple/instawow',
                lambda: aiohttp.web.Response(status=500),
                single_use=True,
            )
        )
        assert await is_outdated() == (False, '0.1.0')

    # Endpoint responsive, version not cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '0.1.0')
        iw_add_routes(
            Route(r'//pypi\.org/simple/instawow', {'versions': ['1.0.0']}, single_use=True)
        )
        assert await is_outdated() == (True, '1.0.0')

    # version == '0+dev', version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '0+dev')
        assert await is_outdated() == (False, '')

    # Endpoint not responsive, version cached
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '0.1.0')
        iw_add_routes(
            Route(
                r'//pypi\.org/simple/instawow',
                lambda: aiohttp.web.Response(status=500),
                single_use=True,
            )
        )
        assert await is_outdated() == (True, '1.0.0')

    # Endpoint responsive, version cached and version same
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '0.1.0')
        iw_add_routes(
            Route(r'//pypi\.org/simple/instawow', {'versions': ['1.0.0']}, single_use=True)
        )
        assert await is_outdated() == (True, '1.0.0')

    # Endpoint responsive, version cached and version different
    with monkeypatch.context() as patcher:
        patcher.setattr('instawow._version.get_version', lambda: '1.0.0')
        iw_add_routes(
            Route(r'//pypi\.org/simple/instawow', {'versions': ['1.0.0']}, single_use=True)
        )
        assert await is_outdated() == (False, '1.0.0')



from __future__ import annotations

import pytest

from instawow import config_ctx
from instawow.definitions import Defn
from instawow.matchers import (
    AddonFolder,
    Matcher,
    _match_addon_names_with_folder_names,
    _match_folder_name_subsets,
    _match_toc_source_ids,
    get_unreconciled_folders,
)
from instawow.wow_installations import Track, to_flavour

pytestmark = pytest.mark.usefixtures('_iw_config_ctx', '_iw_web_client_ctx')


def write_addons(
    *addons: str,
):
    config = config_ctx.config()

    for addon in addons:
        (config.addon_dir / addon).mkdir()
        (config.addon_dir / addon / f'{addon}.toc').touch()


def write_masque_addon():
    masque = config_ctx.config().addon_dir / 'Masque'
    masque.mkdir()
    (masque / 'Masque.toc').write_text(
        """\
## X-Curse-Project-ID: 13592
## X-WoWI-ID: 12097
""",
        encoding='utf-8',
    )
    return masque


async def test_can_extract_defns_from_addon_folder_toc():
    addon_folder = AddonFolder.from_path(
        to_flavour(config_ctx.config().track), write_masque_addon()
    )
    assert addon_folder
    assert addon_folder.get_defns_from_toc_keys(
        config_ctx.resolvers().addon_toc_key_and_id_pairs
    ) == {Defn('curse', '13592'), Defn('wowi', '12097')}


async def test_reconcile_invalid_addons_discarded():
    config = config_ctx.config()
    folders = get_unreconciled_folders()

    config.addon_dir.joinpath('foo').mkdir()
    config.addon_dir.joinpath('bar').touch()

    assert folders == frozenset()
    assert await _match_toc_source_ids(folders) == []
    assert await _match_folder_name_subsets(folders) == []


@pytest.mark.parametrize(
    ('test_func', 'expected_defns'),
    [
        (
            _match_toc_source_ids,
            {
                Defn('curse', '13592'),
                Defn('wowi', '12097'),
                Defn('github', '44074003'),
                Defn('wago', 'kRNLgpGo'),
            },
        ),
        (
            _match_folder_name_subsets,
            {
                Defn('curse', '13592'),
                Defn('wowi', '12097'),
            },
        ),
        (
            _match_addon_names_with_folder_names,
            {
                Defn('curse', '13592'),
                Defn('wowi', '12097'),
                Defn('github', '44074003'),
            },
        ),
    ],
)
async def test_reconcile_multiple_defns_per_addon_contained_in_results(
    test_func: Matcher,
    expected_defns: set[Defn],
):
    write_masque_addon()
    ((_, matches),) = await test_func(get_unreconciled_folders())
    assert expected_defns == set(matches)


@pytest.mark.parametrize(
    ('iw_profile_config_values', 'expected_defns'),
    [
        (
            Track.Retail,
            {
                Defn('wowi', '12097'),
                Defn('curse', '13592'),
            },
        ),
        (
            Track.Classic,
            {
                Defn('curse', '13592'),
            },
        ),
        (
            Track.VanillaClassic,
            {
                Defn('wowi', '12097'),
                Defn('curse', '13592'),
            },
        ),
    ],
    indirect=['iw_profile_config_values'],
)
async def test_reconcile_results_vary_by_game_flavour(
    expected_defns: set[Defn],
):
    write_addons('Masque')
    ((_, matches),) = await _match_folder_name_subsets(get_unreconciled_folders())
    assert expected_defns == set(matches)



from __future__ import annotations

import os
import ssl
from collections.abc import AsyncIterator
from contextlib import AsyncExitStack, asynccontextmanager
from functools import partial
from typing import Any, NotRequired, Protocol, TypedDict

import aiohttp
import aiohttp_client_cache
import aiohttp_client_cache.session

type CachedSession = aiohttp_client_cache.session.CachedSession


_USER_AGENT = 'instawow (+https://github.com/layday/instawow)'

_DEFAULT_EXPIRE = 0  # Do not cache by default.
CACHE_INDEFINITELY = -1

_PROGRESS_TICK_INTERVAL = 0.1


class _TraceConfigCtx[TraceRequestCtxT](Protocol):  # pragma: no cover
    trace_request_ctx: TraceRequestCtxT | None


@asynccontextmanager
async def _setup_progress_tracker():
    import asyncio

    from .._utils.aio import cancel_tasks
    from ..progress_reporting import Progress, get_next_progress_id, update_progress

    progress_tickers = set[asyncio.Task[None]]()

    async def do_on_request_end(
        _client_session: aiohttp.ClientSession,
        trace_config_ctx: Any,
        params: aiohttp.TraceRequestEndParams,
        /,
    ):
        trace_config_ctx_: _TraceConfigCtx[TypedDict[{'progress': NotRequired[Progress[Any]]}]] = (
            trace_config_ctx
        )
        progress = (trace_config_ctx_.trace_request_ctx or {}).get('progress')
        if progress:

            async def do_update_progress():
                response = params.response
                progress_id = get_next_progress_id()
                progress['total'] = total = (
                    None
                    if aiohttp.hdrs.CONTENT_ENCODING in response.headers
                    else response.content_length
                )
                update_progress(progress_id, progress)

                try:
                    if total is None:
                        await response.content.wait_eof()
                    else:
                        while not response.content.is_eof():
                            progress['current'] = response.content.total_bytes
                            update_progress(progress_id, progress)

                            await asyncio.sleep(_PROGRESS_TICK_INTERVAL)

                finally:
                    update_progress(progress_id, 'unset')

            task = asyncio.create_task(do_update_progress())
            progress_tickers.add(task)
            task.add_done_callback(progress_tickers.remove)

    trace_config = aiohttp.TraceConfig()
    trace_config.on_request_end.append(do_on_request_end)
    trace_config.freeze()

    try:
        yield trace_config
    finally:
        await cancel_tasks(progress_tickers)


def get_ssl_context(cloudflare_compat: bool = False) -> ssl.SSLContext:
    import truststore

    ssl_context = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    if cloudflare_compat:
        ssl_context.minimum_version = ssl.TLSVersion.TLSv1_3

    return ssl_context


@asynccontextmanager
async def init_web_client(
    parent_dir: os.PathLike[str] | None, *, with_progress: bool = False
) -> AsyncIterator[CachedSession]:
    make_client_session = partial(
        aiohttp_client_cache.session.CachedSession,
        connector=aiohttp.TCPConnector(limit_per_host=20, ssl=get_ssl_context()),
        headers={'User-Agent': _USER_AGENT},
        timeout=aiohttp.ClientTimeout(connect=60, sock_connect=10, sock_read=20),
        trust_env=True,  # Respect the ``http(s)_proxy`` env var
    )

    async with AsyncExitStack() as async_exit_stack:
        if with_progress:
            progress_trace_config = await async_exit_stack.enter_async_context(
                _setup_progress_tracker()
            )
            make_client_session = partial(
                make_client_session, trace_configs=[progress_trace_config]
            )

        cache_backend = aiohttp_client_cache.CacheBackend(
            allowed_codes=(
                200,
                206,  # Partial Content - returned for successful range requests
                501,  # Not Implemented - returned by GH when a range is out of bounds
            ),
            allowed_methods=('GET', 'POST'),
            expire_after=_DEFAULT_EXPIRE,
            include_headers=True,
        )
        if parent_dir is None:
            cache_backend.disabled = True
        else:
            from ._cache import make_cache

            (
                cache_backend.responses,
                cache_backend.redirects,
            ) = await async_exit_stack.enter_async_context(make_cache(parent_dir))

        client_session = await async_exit_stack.enter_async_context(
            make_client_session(cache=cache_backend)
        )
        yield client_session



from __future__ import annotations

import inspect
from collections.abc import Awaitable, Callable, Collection, Set
from functools import wraps
from typing import Literal, LiteralString, Protocol, TypedDict, overload

from typing_extensions import TypeIs

from .definitions import Strategies, Strategy
from .pkg_db import models as pkg_models


class Result[StatusT: LiteralString](Protocol):  # pragma: no cover
    status: StatusT

    def __str__(self) -> str: ...


class _SuccessResult(Result[Literal['success']]):
    status = 'success'


class PkgInstalled(_SuccessResult):
    def __init__(self, pkg: pkg_models.Pkg, *, dry_run: bool = False) -> None:
        super().__init__()
        self.pkg = pkg
        self.dry_run = dry_run

    def __str__(self) -> str:
        return f'{"would have installed" if self.dry_run else "installed"} {self.pkg.version}'


class PkgUpdated(_SuccessResult):
    def __init__(
        self, old_pkg: pkg_models.Pkg, new_pkg: pkg_models.Pkg, *, dry_run: bool = False
    ) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg
        self.dry_run = dry_run

    def __str__(self) -> str:
        message = (
            f'{"would have updated" if self.dry_run else "updated"}'
            f' {self.old_pkg.version} to {self.new_pkg.version}'
        )

        if self.old_pkg.slug != self.new_pkg.slug:
            message += f' with new slug {self.new_pkg.slug!r}'

        if self.old_pkg.options != self.new_pkg.options:
            old_strategies = self.old_pkg.to_defn().strategies
            new_strategies = self.new_pkg.to_defn().strategies
            message += ' with new strategies: '
            message += '; '.join(
                f'{s}={v!r}' for s, v in new_strategies.items() - old_strategies.items()
            )

        return message


class PkgRemoved(_SuccessResult):
    def __init__(self, old_pkg: pkg_models.Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg

    def __str__(self) -> str:
        return 'removed'


class ManagerError(Result[Literal['failure']], Exception):
    status = 'failure'

    def __str__(self) -> str:
        raise NotImplementedError


class PkgAlreadyInstalled(ManagerError):
    def __str__(self) -> str:
        return 'package already installed'


class PkgConflictsWithInstalled(ManagerError):
    def __init__(
        self, conflicting_pkgs: Collection[TypedDict[{'source': str, 'id': str, 'name': str}]]
    ) -> None:
        super().__init__()
        self.conflicting_pkgs = conflicting_pkgs

    def __str__(self) -> str:
        return (
            'package folders conflict with installed package'
            + ('s ' if len(self.conflicting_pkgs) > 1 else ' ')
            + ', '.join(f'{c["name"]} ({c["source"]}:{c["id"]})' for c in self.conflicting_pkgs)
        )


class PkgConflictsWithUnreconciled(ManagerError):
    def __init__(self, folders: Set[str]) -> None:
        super().__init__()
        self.folders = folders

    def __str__(self) -> str:
        folders = ', '.join(f"'{f}'" for f in self.folders)
        return f'package folders conflict with {folders}'


class PkgNonexistent(ManagerError):
    def __str__(self) -> str:
        return 'package does not exist'


class PkgFilesMissing(ManagerError):
    def __init__(self, reason: str = 'no files are available for download') -> None:
        super().__init__()
        self._reason = reason

    def __str__(self) -> str:
        return self._reason


class PkgFilesNotMatching(ManagerError):
    def __init__(self, strategies: Strategies) -> None:
        super().__init__()
        self.strategies = strategies

    def __str__(self) -> str:
        return 'no files found for: ' + '; '.join(f'{s}={v!r}' for s, v in self.strategies.items())


class PkgNotInstalled(ManagerError):
    def __str__(self) -> str:
        return 'package is not installed'


class PkgSourceInvalid(ManagerError):
    def __str__(self) -> str:
        return 'package source is invalid'


class PkgSourceDisabled(ManagerError):
    def __init__(self, reason: str | None = None) -> None:
        super().__init__()
        self._reason = reason

    def __str__(self) -> str:
        return f'package source is disabled{f": {self._reason}" if self._reason else ""}'


class PkgUpToDate(ManagerError):
    def __init__(self, is_pinned: bool) -> None:
        super().__init__()
        self.is_pinned = is_pinned

    def __str__(self) -> str:
        return f'package is {"pinned" if self.is_pinned else "up to date"}'


class PkgStrategiesUnsupported(ManagerError):
    def __init__(self, strategies: Collection[Strategy]) -> None:
        super().__init__()
        self.strategies = strategies

    def __str__(self) -> str:
        return f'strategies are not valid for source: {", ".join(sorted(self.strategies))}'


class InternalError(Result[Literal['error']], Exception):
    status = 'error'

    def __str__(self) -> str:
        return f'internal error: "{self.args[0]}"'


type AnyResult[T] = T | ManagerError | InternalError


def _handle_internal_error(error: BaseException):
    from ._logging import logger

    traceback = error.__traceback__
    if traceback is not None:
        error = error.with_traceback(traceback.tb_next)

    logger.opt(
        exception=error,
    ).error('Unclassed error')

    return InternalError(error)


@overload
def resultify[**P, T](fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[AnyResult[T]]]: ...
@overload
def resultify[**P, T](fn: Callable[P, T]) -> Callable[P, AnyResult[T]]: ...


def resultify[**P](fn: Callable[P, object]):  # pyright: ignore[reportInconsistentOverload]
    "Capture raw errors and wrap them around ``InternalError``."

    if inspect.iscoroutinefunction(fn):

        @wraps(fn)
        async def async_wrapper(*args: P.args, **kwargs: P.kwargs):
            try:
                return await fn(*args, **kwargs)

            except (ManagerError, InternalError) as exception:
                return exception

            except BaseException as error:
                return _handle_internal_error(error)

        return async_wrapper

    else:

        @wraps(fn)
        def sync_wrapper(*args: P.args, **kwargs: P.kwargs):
            try:
                return fn(*args, **kwargs)

            except (ManagerError, InternalError) as exception:
                return exception

            except BaseException as error:
                return _handle_internal_error(error)

        return sync_wrapper


def is_error_result(result: AnyResult[object]) -> TypeIs[ManagerError | InternalError]:
    return isinstance(result, (ManagerError, InternalError))
